import numpy as np
import matplotlib.pyplot as plt
from scipy import stats
from scipy import optimize

i = complex(0.0, 1.0) # Assigning i=sqrt(-1)

class rough_Heston: 
    def __init__(self, time_steps=100, alpha=0.6, r=0.0, rho=-0.7, eta=0.3, labda=0.3, theta=0.02, V0=0.02, n=20, rn=2.5):
        self.time_steps = time_steps
        self.r = r
        self.rho = rho
        self.eta = eta
        self.labda = labda
        self.theta = theta
        self.V0 = V0
        self.n = n
        self.c = (rn**(1 - alpha) - 1) * rn**((alpha - 1) * (1 + n / 2)) * rn**((1 - alpha) * np.linspace(1.0, n, n)) / (np.math.gamma(alpha) * np.math.gamma(2 - alpha))
        self.gamma = (1 - alpha) / (2 - alpha) * (rn**(2 - alpha) - 1) / (rn**(1 - alpha) -1) * rn**(np.linspace(1.0, n, n) - 1 - n / 2)
        
    def gn(self, t):
        return self.V0 + self.labda * self.theta * np.sum(self.c / self.gamma * (1 - np.exp(-self.gamma * t)))

    def F(self, u, v):
        return 0.5 * (u**2 - u) + (self.rho * self.eta * u - self.labda) * v + 0.5 * self.eta**2 * v**2

    def Riccati_solver(self, u):
        psi = np.zeros((self.time_steps+1, self.n), dtype=complex)
        F_u = np.zeros(self.time_steps+1, dtype=complex)
        
        for j in range(self.time_steps+1):
            F_u[j] = self.F(u, sum(self.c * psi[j-1]))
            psi[j] = 1 / (1 + self.gamma * self.h) * (psi[j-1] + self.h * F_u[j])
            
        return psi, F_u
    
    def trapezoidal_convolution(self, x, y):
        return (sum(x * y[::-1]) - 0.5 * (x[0] * y[-1] + x[-1] * y[0])) * self.h
    
    def characteristic(self, t, u):
        if u == 0:
            return 1
        
        self.h = t / self.time_steps
        g_0 = np.zeros(self.time_steps+1)
        
        for j in range(self.time_steps+1):
            g_0[j] = self.gn(j*self.h)
            
        _, F_u = self.Riccati_solver(u*i)
        return np.exp(self.r * i * u * t + self.trapezoidal_convolution(F_u, g_0))
    
class Black_Scholes:
    def __init__(self, r=0.0, sigma=0.25, CP=1):
        self.r = r
        self.sigma = sigma
        self.CP = CP

    def cF(self, t, u):
        if u == 0:
            return 1
        else:
            return np.exp(i * u * (self.r - 0.5 * self.sigma**2) * t - 0.5 * self.sigma**2 * u**2 * t)
    
    # Black-Scholes European option price
    def OptionPrice(self, S, t, sigma=None):
        if sigma == None:
            sigma = self.sigma
            
        d1 = (np.log(S) + (self.r + sigma**2 / 2) * t) / (sigma * np.sqrt(t))
        d2 = d1 - (sigma * np.sqrt(t))
        
        if self.CP == 1:
            return S * stats.norm.cdf(d1) - np.exp(-self.r * t) * stats.norm.cdf(d2)
        if self.CP == -1:
            return np.exp(-self.r * t) * stats.norm.cdf(-d2) - S * stats.norm.cdf(-d1)
        
    # Finite difference method for an American Option
    def AmericanFD(self, T, Nsteps=10000, S_high=2.0):
        dt = T / Nsteps  # Time grid
        dx = S_high / 2 * np.sqrt(dt) # Space grid
        S = np.arange(dx, S_high + dx, dx)
        Ndx = len(S) - 1
        x_int = np.arange(1, Ndx)     
        
        # American option price grid
        price = np.nan * np.ones([Ndx+1, Nsteps+1])      # Put price at each time step
        price[:,-1] = np.maximum(self.CP * (S - 1.0), 0) # Set payoff at maturity
        hc = np.nan * np.ones(Ndx + 1)

        # Step backwards through time to compute continuation values and compare with intrinsic values
        for n in np.arange(Nsteps, 0, -1):             
            hc[x_int] =  price[x_int,n] + dt * (-self.r * price[x_int,n] + self.r * S[x_int] * (price[x_int+1,n] - price[x_int-1,n]) / (2 * dx) + 0.5 * self.sigma**2 * S[x_int]**2 * (price[x_int+1,n] - 2 * price[x_int,n] + price[x_int-1,n]) / (dx)**2)
            hc[0] = 2 * hc[1] - hc[2]
            hc[Ndx] = 2 * hc[Ndx-1] - hc[Ndx-2]
            price[:,n-1] = np.maximum(hc, self.CP * (S - 1.0))
        
        return S, price
    
    def implied_volatility(self, S, option_price, t):
        equation = lambda sigma: self.OptionPrice(S, t, sigma) - option_price
        return optimize.newton(equation, self.sigma) # Initial guess for implied volatility

def Heston_cF(t, u, V0, r=0.0, rho=0.0, eta=0.1, labda=2.0, theta=0.01, sigma=1.0):
    if u == 0:
        return 1
    M = np.sqrt((sigma * rho * eta * u * i - labda)**2 + sigma**2 * eta**2 * (u * i +  u**2))
    N = (rho * sigma * eta * i * u - labda - M) / (rho * sigma * eta * i * u - labda + M)
    A = r * i * u * t + labda * theta / eta**2 * (-(rho * sigma * eta * i * u - labda - M) * t - 2 * np.log((1 - N * np.exp(M * t)) / (1 - N)))
    C = ((np.exp(M * t) - 1) * (rho * sigma * eta * i * u - labda - M)) / (eta**2 * (1 - N * np.exp(M * t)))
    return np.exp(A + C * V0)

# COS method for option pricing from characteristic function 
class COS_Mthd:
    def __init__(self, N=700, L=10, r=0.0):
        self.N = N
        self.L = L
        self.r = r
        
    def CallPutOptionPrice(self, cF, CP, S0, tau):     
        # Truncation domain
        a = -self.L * np.sqrt(np.sqrt(tau))
        b = self.L * np.sqrt(np.sqrt(tau))
        x0 = np.log(S0)   
        
        # Summation from k = 0 to k = N - 1
        k = np.linspace(0, self.N-1, self.N)
        u = k * np.pi / (b - a) 
        cf_vector = np.zeros(self.N, dtype='complex')
        
        for j in range(self.N):
            cf_vector[j] = cF(u[j])
            
        H_k = self.CallPutCoefficients(CP, a, b, k) # Determine coefficients for option prices
        temp = cf_vector * H_k
        temp[0] = 0.5 * temp[0]    
        mat = np.exp(i * np.outer((x0 - a) , u))
        return np.exp(-self.r * tau) * np.real(mat.dot(temp))     

    def CallPutCoefficients(self, CP, a, b, k):
        c = 0.5 * (1 - CP) * a
        d = 0.5 * (1 + CP) * b
        Chi_k, Psi_k = self.Chi_Psi(a, b, c, d, k)        
        return 2 / (b - a) * CP * (Chi_k - Psi_k)

    def Chi_Psi(self, a, b, c, d, k):
        psi = np.sin(k * np.pi * (d - a) / (b - a)) - np.sin(k * np.pi * (c - a) / (b - a))
        psi[1:] = psi[1:] * (b - a) / (k[1:] * np.pi)
        psi[0] = d - c
        chi = 1.0 / (1.0 + (k * np.pi / (b - a))**2) 
        expr1 = np.cos(k * np.pi * (d - a) / (b - a)) * np.exp(d)  - np.cos(k * np.pi * (c - a) / (b - a)) * np.exp(c)
        expr2 = k * np.pi / (b - a) * np.sin(k * np.pi * (d - a) / (b - a)) * np.exp(d) - k * np.pi / (b - a) * np.sin(k * np.pi * (c - a) / (b - a)) * np.exp(c)
        chi = chi * (expr1 + expr2)
        return chi, psi

class Plot:
    def __init__(self, matrix, labels, S_plot, name):
        self.matrix = matrix
        self.labels = labels
        self.S_plot = S_plot
        self.name = name
        n_plot, self.N_t, _ = np.shape(self.matrix)
        self.n_plot = min(n_plot, len(self.labels)) # Only plot lines for which both values and label is defined
        self.linestyle = ['-',':','--','-.']
        self.colors = ['b','g','r','c','m','y','k']

    def figures(self, title, n_figures=4, SaveOutput=False):    
        plt.figure(figsize = (12, 2.5 * n_figures)) # Figure options

        for m,j in enumerate([int(x) for x in np.linspace(0,self.N_t-1,n_figures)]):
            plt.subplot(int(np.ceil(n_figures/2)), 2, m+1) # Specify subplot
            
            # Plot option prices of simulated process values
            for k in range(self.n_plot):
                plt.plot(self.S_plot, self.matrix[k,j], color=self.colors[k%len(self.colors)], label=self.labels[k], linestyle=self.linestyle[k%len(self.linestyle)]) 
    
            # Subplot options
            plt.xlim(xmin=np.min(self.S_plot), xmax=np.max(self.S_plot))
            plt.xlabel(r"Moneyness")
            plt.ylabel(r"Option price")
            plt.title(title[j])
            plt.grid(linestyle=':')
    
            if j == 0:
                plt.legend()

        # Save figure
        if SaveOutput:
            plt.savefig('Figures/'+self.name+'.png', bbox_inches='tight')
        
        plt.show()
        plt.close()
        
    # Plot of the error
    def error(self, times, SaveOutput=False, order='2'):
        plt.figure()
        
        for j in range(1, self.n_plot):
            if order == '2':
                error = np.linalg.norm(self.matrix[j] - self.matrix[0], 2, 1) / np.linalg.norm(self.matrix[0], 2, 1)
            elif order == 'max': 
                error = np.linalg.norm(self.matrix[j] - self.matrix[0], np.inf, 1)
            plt.plot(times, error, label = self.labels[j], color=self.colors[j%len(self.colors)], linestyle=self.linestyle[j%len(self.linestyle)])
            
        plt.ylim(ymin=0.0)
        plt.xlim(xmin=0.0, xmax=np.max(times))
        plt.xlabel('t')
        if order == '2':
            plt.ylabel('Relative L2-error')
        elif order == 'max':
            plt.ylabel('Max absolute error')
        plt.legend()
        
        # Save figure
        if SaveOutput:
            plt.savefig('Figures/'+self.name+'_'+order+'error.png')
            
        plt.show()
        plt.close()
        
    # Make an image of each timestep for animation
    def make_animation(self, flip=True):
        for j in range(self.N_t):
            ax = plt.axes(xlim=(np.min(self.S_plot),np.max(self.S_plot)), ylim=(-0.1,2.1))
            ax.set_xlabel('Moneyness')
            ax.set_ylabel('Option price')
            ax.grid(linestyle=':')
            
            for i in range(self.n_plot):
                ax.plot(self.S_plot, self.matrix[i][int(2 * (0.5 - flip) * j + flip * (self.N_t - 1))], lw=2, color=self.colors[i%len(self.colors)], label=self.labels[i])
            
            plt.legend()    
            plt.savefig('Figures/'+self.name+'_animation-'+str(j)+'.png')
            plt.close()